(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2020 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

BeginPackage["AtomicDensityMatrix`ADMPerturbativeSolve`", {"AtomicDensityMatrix`"}];

Begin["`Private`"];

ComplexFrequencyMatrix[h_, g_] :=
    Table[(h[[i, i]] - h[[j, j]]) - I / 2(g[[i, i]] + g[[j, j]]), {i, Length[h]}, {j, Length[h]}];

If[$VersionNumber < 12, DiagonalMatrixQ[expr_] := MatrixQ[expr] && expr === DiagonalMatrix[Tr[expr, List]]];

SteadyStatePerturbativeSolve[v_?MatrixQ, l_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), {m_, n_}, {0}] :=
    -((I * l[[m, n]]) / w[[m, n]]);

SteadyStatePerturbativeSolve[v_?MatrixQ, l_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), {m_, n_}, {1}] :=
    Sum[-((I * v[[p[0], n]] * l[[m, p[0]]]) / (w[[m, n]] * w[[m, p[0]]])) +
        (I * v[[m, p[0]]] * l[[p[0], n]]) / (w[[m, n]] * w[[p[0], n]]),
        {p[0], 1, Length[v]}
    ];

SteadyStatePerturbativeSolve[v_?MatrixQ, l_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), {m_, n_}, {2}] :=
    Module[ {FOelems = Most[ArrayRules[SparseArray[f]][[All, 1]]]},
        Sum[If[ MemberQ[FOelems, {m, n, s[0], r[0]}],
            Sum[(v[[p[2], s[0]]] *
                v[[r[0], p[1]]] * l[[p[1],
                p[2]]]) / (w[[m, n]] * w[[p[1], p[2]]] *
                w[[p[1], s[0]]] * w[[r[0],
                s[0]]]) + (v[[p[1], s[0]]] *
                v[[r[0], p[2]]] * l[[p[2], p[1]]]) /
                (w[[m, n]] * w[[p[2],
                    p[1]]] * w[[r[0], p[1]]] * w[[r[0], s[0]]]),
                {p[2], 1, Length[v]}, {p[1], 1, Length[v]}] *
                f[[m, n, s[0], r[0]]],
            0
        ], {r[0], 1, Length[v]},
            {s[0], 1, Length[v]}] +
            Sum[-((I * v[[p[0], n]] * v[[p[1], p[0]]] * l[[m, p[1]]]) /
                (w[[m, n]] * w[[m, p[0]]] * w[[m,
                    p[1]]])) +
                (I * v[[m, p[0]]] *
                    v[[p[1], n]] * l[[p[0], p[1]]]) / (w[[m,
                    n]] * w[[p[0], n]] *
                    w[[p[0], p[1]]]) - (I * v[[m, p[0]]] *
                v[[p[0], p[1]]] * l[[p[1], n]]) /
                (w[[m, n]] * w[[p[0], n]] * w[[p[1],
                    n]]) + (I * v[[m, p[1]]] *
                v[[p[0], n]] * l[[p[1], p[0]]]) /
                (w[[m, n]] * w[[m, p[0]]] * w[[p[1],
                    p[0]]]), {p[1], 1, Length[v]}, {p[0], 1, Length[v]}]
    ];

SteadyStatePerturbativeSolve[v_?MatrixQ, l_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), {m_, n_}, {3}] :=
    Module[{FOelems = Most[ArrayRules[SparseArray[f]][[All, 1]]]},
        Sum[If[MemberQ[FOelems, {m, n, s[0], r[0]}],

            Sum[(v[[p[2], s[0]]] * v[[p[3], p[2]]] *
                v[[r[0], p[1]]] * l[[p[1], p[3]]]) /
                (w[[m, n]] * w[[p[1],
                    p[2]]] * w[[p[1], p[3]]] * w[[p[1],
                    s[0]]] * w[[r[0], s[0]]]) -
                (v[[p[1], p[2]]] * v[[p[3], s[0]]] *
                    v[[r[0], p[1]]] * l[[p[2], p[3]]]) /
                    (w[[m, n]] * w[[p[1],
                        s[0]]] * w[[p[2], p[3]]] * w[[p[2],
                        s[0]]] * w[[r[0], s[0]]]) -
                (v[[p[1], p[3]]] * v[[p[2], s[0]]] *
                    v[[r[0], p[1]]] * l[[p[3], p[2]]]) /
                    (w[[m, n]] * w[[p[1],
                        p[2]]] * w[[p[1], s[0]]] * w[[p[3],
                        p[2]]] * w[[r[0], s[0]]]) +
                (v[[p[1], s[0]]] * v[[p[3], p[1]]] *
                    v[[r[0], p[2]]] * l[[p[2], p[3]]]) /
                    (w[[m, n]] * w[[p[2],
                        p[1]]] * w[[p[2], p[3]]] * w[[r[0],
                        p[1]]] * w[[r[0], s[0]]]) -
                (v[[p[1], s[0]]] * v[[p[2], p[3]]] *
                    v[[r[0], p[2]]] * l[[p[3], p[1]]]) /
                    (w[[m, n]] * w[[p[2],
                        p[1]]] * w[[p[3], p[1]]] * w[[r[0],
                        p[1]]] * w[[r[0], s[0]]]) +
                (v[[p[1], s[0]]] * v[[p[2], p[1]]] *
                    v[[r[0], p[3]]] * l[[p[3], p[2]]]) /
                    (w[[m, n]] * w[[p[3],
                        p[2]]] * w[[r[0], p[1]]] * w[[r[0],
                        p[2]]] * w[[r[0], s[0]]]),
                {p[3], 1, Length[v]}, {p[2], 1, Length[v]}, {p[1], 1,
                Length[v]}] * f[[m, n, s[0], r[0]]], 0],
            {r[0], 1, Length[v]}, {s[0], 1, Length[v]}] +
            Sum[If[MemberQ[FOelems, {m, p[0], s[1], r[1]}],

                Sum[(v[[p[0], n]] * v[[p[3], s[1]]] *
                    v[[r[1], p[2]]] * l[[p[2], p[3]]]) /
                    (w[[m, n]] * w[[m,
                        p[0]]] * w[[p[2], p[3]]] * w[[p[2],
                        s[1]]] * w[[r[1], s[1]]]) +
                    (v[[p[0], n]] * v[[p[2], s[1]]] *
                        v[[r[1], p[3]]] * l[[p[3],
                        p[2]]]) / (w[[m, n]] * w[[m, p[0]]] *
                        w[[p[3], p[2]]] * w[[r[1],
                        p[2]]] * w[[r[1], s[1]]]), {p[3], 1, Length[v]},
                    {p[2], 1, Length[v]}] * f[[m, p[0], s[1], r[1]]], 0] +
                If[MemberQ[FOelems, {p[0], n, s[1], r[1]}],

                    Sum[-((v[[m, p[0]]] * v[[p[3], s[1]]] *
                        v[[r[1], p[2]]] * l[[p[2], p[3]]]) /
                        (w[[m, n]] * w[[p[0],
                            n]] * w[[p[2], p[3]]] * w[[p[2],
                            s[1]]] * w[[r[1], s[1]]])) -
                        (v[[m, p[0]]] * v[[p[2], s[1]]] *
                            v[[r[1], p[3]]] * l[[p[3],
                            p[2]]]) / (w[[m, n]] * w[[p[0], n]] *
                            w[[p[3], p[2]]] * w[[r[1],
                            p[2]]] * w[[r[1], s[1]]]), {p[3], 1, Length[v]},
                        {p[2], 1, Length[v]}] * f[[p[0], n, s[1], r[1]]],
                    0], {r[1], 1, Length[v]},
                {s[1], 1, Length[v]}, {p[0], 1, Length[v]}] +
            Sum[-((I * v[[p[0], n]] * v[[p[1], p[0]]] *
                v[[p[2], p[1]]] * l[[m, p[2]]]) / (w[[m,
                n]] * w[[m, p[0]]] *
                w[[m, p[1]]] * w[[m, p[2]]])) + (I *
                v[[m, p[0]]] * v[[p[1], n]] * v[[p[2], p[1]]] *
                l[[p[0], p[2]]]) / (w[[m,
                n]] * w[[p[0], n]] * w[[p[0],
                p[1]]] * w[[p[0], p[2]]]) -
                (I * v[[m, p[0]]] * v[[p[0], p[1]]] *
                    v[[p[2], n]] * l[[p[1], p[2]]]) / (w[[m,
                    n]] * w[[p[0], n]] *
                    w[[p[1], n]] * w[[p[1], p[2]]]) + (I *
                v[[m, p[1]]] * v[[p[0], n]] * v[[p[2], p[0]]] *
                l[[p[1], p[2]]]) / (w[[m,
                n]] * w[[m, p[0]]] * w[[p[1],
                p[0]]] * w[[p[1], p[2]]]) +
                (I * v[[m, p[0]]] * v[[p[0], p[1]]] *
                    v[[p[1], p[2]]] * l[[p[2], n]]) / (w[[m,
                    n]] * w[[p[0], n]] *
                    w[[p[1], n]] * w[[p[2], n]]) - (I *
                v[[m, p[1]]] * v[[p[0], n]] * v[[p[1], p[2]]] *
                l[[p[2], p[0]]]) / (w[[m,
                n]] * w[[m, p[0]]] * w[[p[1],
                p[0]]] * w[[p[2], p[0]]]) +
                (I * v[[m, p[2]]] * v[[p[0], n]] *
                    v[[p[1], p[0]]] * l[[p[2], p[1]]]) / (w[[m,
                    n]] * w[[m, p[0]]] *
                    w[[m, p[1]]] * w[[p[2], p[1]]]) - (I *
                v[[m, p[0]]] * v[[p[0], p[2]]] * v[[p[1], n]] *
                l[[p[2], p[1]]]) / (w[[m,
                n]] * w[[p[0], n]] * w[[p[0],
                p[1]]] * w[[p[2], p[1]]]),
                {p[2], 1, Length[v]}, {p[1], 1, Length[v]}, {p[0], 1,
                Length[v]}]];

SteadyStatePerturbativeSolve[v_?MatrixQ, l_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), {m_, n_}, {4}] :=
    Module[{FOelems = Most[ArrayRules[SparseArray[f]][[All, 1]]]},
        Sum[(Sum[If[MemberQ[FOelems, {p[1], p[2], s[3], r[3]}],
            Sum[(I * v[[p[2], s[0]]] * v[[p[5], s[3]]] * v[[r[0], p[1]]] *
                v[[r[3], p[4]]] * l[[p[4], p[5]]]) / (w[[m, n]] *
                w[[p[1], p[2]]] * w[[p[1], s[0]]] * w[[p[4], p[5]]] *
                w[[p[4], s[3]]] * w[[r[0], s[0]]] * w[[r[3], s[3]]]),
                {p[5], 1, Length[v]}, {p[4], 1, Length[v]}] *
                f[[p[1], p[2], s[3], r[3]]], 0] +
            If[MemberQ[FOelems, {p[1], p[2], s[3], r[3]}],
                Sum[(I * v[[p[2], s[0]]] * v[[p[4], s[3]]] * v[[r[0], p[1]]] *
                    v[[r[3], p[5]]] * l[[p[5], p[4]]]) / (w[[m, n]] *
                    w[[p[1], p[2]]] * w[[p[1], s[0]]] * w[[p[5], p[4]]] *
                    w[[r[0], s[0]]] * w[[r[3], p[4]]] * w[[r[3], s[3]]]),
                    {p[5], 1, Length[v]}, {p[4], 1, Length[v]}] *
                    f[[p[1], p[2], s[3], r[3]]], 0] +
            If[MemberQ[FOelems, {p[2], p[1], s[3], r[3]}],
                Sum[(I * v[[p[1], s[0]]] * v[[p[5], s[3]]] * v[[r[0], p[2]]] *
                    v[[r[3], p[4]]] * l[[p[4], p[5]]]) / (w[[m, n]] *
                    w[[p[2], p[1]]] * w[[p[4], p[5]]] * w[[p[4], s[3]]] *
                    w[[r[0], p[1]]] * w[[r[0], s[0]]] * w[[r[3], s[3]]]),
                    {p[5], 1, Length[v]}, {p[4], 1, Length[v]}] *
                    f[[p[2], p[1], s[3], r[3]]], 0] +
            If[MemberQ[FOelems, {p[2], p[1], s[3], r[3]}],
                Sum[(I * v[[p[1], s[0]]] * v[[p[4], s[3]]] * v[[r[0], p[2]]] *
                    v[[r[3], p[5]]] * l[[p[5], p[4]]]) / (w[[m, n]] *
                    w[[p[2], p[1]]] * w[[p[5], p[4]]] * w[[r[0], p[1]]] *
                    w[[r[0], s[0]]] * w[[r[3], p[4]]] * w[[r[3], s[3]]]),
                    {p[5], 1, Length[v]}, {p[4], 1, Length[v]}] *
                    f[[p[2], p[1], s[3], r[3]]], 0], {r[3], 1, Length[v]},
            {s[3], 1, Length[v]}, {p[2], 1, Length[v]},
            {p[1], 1, Length[v]}] +
            Sum[(v[[p[2], s[0]]] * v[[p[3], p[2]]] * v[[p[4], p[3]]] *
                v[[r[0], p[1]]] * l[[p[1], p[4]]]) / (w[[m, n]] *
                w[[p[1], p[2]]] * w[[p[1], p[3]]] * w[[p[1], p[4]]] *
                w[[p[1], s[0]]] * w[[r[0], s[0]]]) -
                (v[[p[1], p[2]]] * v[[p[3], s[0]]] * v[[p[4], p[3]]] * v[[r[0], p[1]]] *
                    l[[p[2], p[4]]]) / (w[[m, n]] * w[[p[1], s[0]]] *
                    w[[p[2], p[3]]] * w[[p[2], p[4]]] * w[[p[2], s[0]]] *
                    w[[r[0], s[0]]]) - (v[[p[1], p[3]]] * v[[p[2], s[0]]] *
                v[[p[4], p[2]]] * v[[r[0], p[1]]] * l[[p[3], p[4]]]) /
                (w[[m, n]] * w[[p[1], p[2]]] * w[[p[1], s[0]]] *
                    w[[p[3], p[2]]] * w[[p[3], p[4]]] * w[[r[0], s[0]]]) +
                (v[[p[1], p[2]]] * v[[p[2], p[3]]] * v[[p[4], s[0]]] * v[[r[0], p[1]]] *
                    l[[p[3], p[4]]]) / (w[[m, n]] * w[[p[1], s[0]]] *
                    w[[p[2], s[0]]] * w[[p[3], p[4]]] * w[[p[3], s[0]]] *
                    w[[r[0], s[0]]]) + (v[[p[1], p[3]]] * v[[p[2], s[0]]] *
                v[[p[3], p[4]]] * v[[r[0], p[1]]] * l[[p[4], p[2]]]) /
                (w[[m, n]] * w[[p[1], p[2]]] * w[[p[1], s[0]]] *
                    w[[p[3], p[2]]] * w[[p[4], p[2]]] * w[[r[0], s[0]]]) -
                (v[[p[1], p[4]]] * v[[p[2], s[0]]] * v[[p[3], p[2]]] * v[[r[0], p[1]]] *
                    l[[p[4], p[3]]]) / (w[[m, n]] * w[[p[1], p[2]]] *
                    w[[p[1], p[3]]] * w[[p[1], s[0]]] * w[[p[4], p[3]]] *
                    w[[r[0], s[0]]]) + (v[[p[1], p[2]]] * v[[p[2], p[4]]] *
                v[[p[3], s[0]]] * v[[r[0], p[1]]] * l[[p[4], p[3]]]) /
                (w[[m, n]] * w[[p[1], s[0]]] * w[[p[2], p[3]]] *
                    w[[p[2], s[0]]] * w[[p[4], p[3]]] * w[[r[0], s[0]]]) +
                (v[[p[1], s[0]]] * v[[p[3], p[1]]] * v[[p[4], p[3]]] * v[[r[0], p[2]]] *
                    l[[p[2], p[4]]]) / (w[[m, n]] * w[[p[2], p[1]]] *
                    w[[p[2], p[3]]] * w[[p[2], p[4]]] * w[[r[0], p[1]]] *
                    w[[r[0], s[0]]]) - (v[[p[1], s[0]]] * v[[p[2], p[3]]] *
                v[[p[4], p[1]]] * v[[r[0], p[2]]] * l[[p[3], p[4]]]) /
                (w[[m, n]] * w[[p[2], p[1]]] * w[[p[3], p[1]]] *
                    w[[p[3], p[4]]] * w[[r[0], p[1]]] * w[[r[0], s[0]]]) +
                (v[[p[1], s[0]]] * v[[p[2], p[3]]] * v[[p[3], p[4]]] * v[[r[0], p[2]]] *
                    l[[p[4], p[1]]]) / (w[[m, n]] * w[[p[2], p[1]]] *
                    w[[p[3], p[1]]] * w[[p[4], p[1]]] * w[[r[0], p[1]]] *
                    w[[r[0], s[0]]]) - (v[[p[1], s[0]]] * v[[p[2], p[4]]] *
                v[[p[3], p[1]]] * v[[r[0], p[2]]] * l[[p[4], p[3]]]) /
                (w[[m, n]] * w[[p[2], p[1]]] * w[[p[2], p[3]]] *
                    w[[p[4], p[3]]] * w[[r[0], p[1]]] * w[[r[0], s[0]]]) +
                (v[[p[1], s[0]]] * v[[p[2], p[1]]] * v[[p[4], p[2]]] * v[[r[0], p[3]]] *
                    l[[p[3], p[4]]]) / (w[[m, n]] * w[[p[3], p[2]]] *
                    w[[p[3], p[4]]] * w[[r[0], p[1]]] * w[[r[0], p[2]]] *
                    w[[r[0], s[0]]]) - (v[[p[1], s[0]]] * v[[p[2], p[1]]] *
                v[[p[3], p[4]]] * v[[r[0], p[3]]] * l[[p[4], p[2]]]) /
                (w[[m, n]] * w[[p[3], p[2]]] * w[[p[4], p[2]]] *
                    w[[r[0], p[1]]] * w[[r[0], p[2]]] * w[[r[0], s[0]]]) +
                (v[[p[1], s[0]]] * v[[p[2], p[1]]] * v[[p[3], p[2]]] * v[[r[0], p[4]]] *
                    l[[p[4], p[3]]]) / (w[[m, n]] * w[[p[4], p[3]]] *
                    w[[r[0], p[1]]] * w[[r[0], p[2]]] * w[[r[0], p[3]]] *
                    w[[r[0], s[0]]]), {p[4], 1, Length[v]},
                {p[3], 1, Length[v]}, {p[2], 1, Length[v]},
                {p[1], 1, Length[v]}]) * f[[m, n, s[0], r[0]]],
            {r[0], 1, Length[v]}, {s[0], 1, Length[v]}] +
            Sum[If[MemberQ[FOelems, {m, p[0], s[1], r[1]}],
                Sum[(v[[p[0], n]] * v[[p[3], s[1]]] * v[[p[4], p[3]]] * v[[r[1], p[2]]] *
                    l[[p[2], p[4]]]) / (w[[m, n]] * w[[m, p[0]]] *
                    w[[p[2], p[3]]] * w[[p[2], p[4]]] * w[[p[2], s[1]]] *
                    w[[r[1], s[1]]]) - (v[[p[0], n]] * v[[p[2], p[3]]] *
                    v[[p[4], s[1]]] * v[[r[1], p[2]]] * l[[p[3], p[4]]]) /
                    (w[[m, n]] * w[[m, p[0]]] * w[[p[2], s[1]]] *
                        w[[p[3], p[4]]] * w[[p[3], s[1]]] * w[[r[1], s[1]]]) -
                    (v[[p[0], n]] * v[[p[2], p[4]]] * v[[p[3], s[1]]] * v[[r[1], p[2]]] *
                        l[[p[4], p[3]]]) / (w[[m, n]] * w[[m, p[0]]] *
                        w[[p[2], p[3]]] * w[[p[2], s[1]]] * w[[p[4], p[3]]] *
                        w[[r[1], s[1]]]) + (v[[p[0], n]] * v[[p[2], s[1]]] *
                    v[[p[4], p[2]]] * v[[r[1], p[3]]] * l[[p[3], p[4]]]) /
                    (w[[m, n]] * w[[m, p[0]]] * w[[p[3], p[2]]] *
                        w[[p[3], p[4]]] * w[[r[1], p[2]]] * w[[r[1], s[1]]]) -
                    (v[[p[0], n]] * v[[p[2], s[1]]] * v[[p[3], p[4]]] * v[[r[1], p[3]]] *
                        l[[p[4], p[2]]]) / (w[[m, n]] * w[[m, p[0]]] *
                        w[[p[3], p[2]]] * w[[p[4], p[2]]] * w[[r[1], p[2]]] *
                        w[[r[1], s[1]]]) + (v[[p[0], n]] * v[[p[2], s[1]]] *
                    v[[p[3], p[2]]] * v[[r[1], p[4]]] * l[[p[4], p[3]]]) /
                    (w[[m, n]] * w[[m, p[0]]] * w[[p[4], p[3]]] *
                        w[[r[1], p[2]]] * w[[r[1], p[3]]] * w[[r[1], s[1]]]),
                    {p[4], 1, Length[v]}, {p[3], 1, Length[v]},
                    {p[2], 1, Length[v]}] * f[[m, p[0], s[1], r[1]]], 0] +
                If[MemberQ[FOelems, {p[0], n, s[1], r[1]}],
                    Sum[-((v[[m, p[0]]] * v[[p[3], s[1]]] * v[[p[4], p[3]]] * v[[r[1], p[2]]] *
                        l[[p[2], p[4]]]) / (w[[m, n]] * w[[p[0], n]] *
                        w[[p[2], p[3]]] * w[[p[2], p[4]]] * w[[p[2], s[1]]] *
                        w[[r[1], s[1]]])) + (v[[m, p[0]]] * v[[p[2], p[3]]] *
                        v[[p[4], s[1]]] * v[[r[1], p[2]]] * l[[p[3], p[4]]]) /
                        (w[[m, n]] * w[[p[0], n]] * w[[p[2], s[1]]] *
                            w[[p[3], p[4]]] * w[[p[3], s[1]]] * w[[r[1], s[1]]]) +
                        (v[[m, p[0]]] * v[[p[2], p[4]]] * v[[p[3], s[1]]] * v[[r[1], p[2]]] *
                            l[[p[4], p[3]]]) / (w[[m, n]] * w[[p[0], n]] *
                            w[[p[2], p[3]]] * w[[p[2], s[1]]] * w[[p[4], p[3]]] *
                            w[[r[1], s[1]]]) - (v[[m, p[0]]] * v[[p[2], s[1]]] *
                        v[[p[4], p[2]]] * v[[r[1], p[3]]] * l[[p[3], p[4]]]) /
                        (w[[m, n]] * w[[p[0], n]] * w[[p[3], p[2]]] *
                            w[[p[3], p[4]]] * w[[r[1], p[2]]] * w[[r[1], s[1]]]) +
                        (v[[m, p[0]]] * v[[p[2], s[1]]] * v[[p[3], p[4]]] * v[[r[1], p[3]]] *
                            l[[p[4], p[2]]]) / (w[[m, n]] * w[[p[0], n]] *
                            w[[p[3], p[2]]] * w[[p[4], p[2]]] * w[[r[1], p[2]]] *
                            w[[r[1], s[1]]]) - (v[[m, p[0]]] * v[[p[2], s[1]]] *
                        v[[p[3], p[2]]] * v[[r[1], p[4]]] * l[[p[4], p[3]]]) /
                        (w[[m, n]] * w[[p[0], n]] * w[[p[4], p[3]]] *
                            w[[r[1], p[2]]] * w[[r[1], p[3]]] * w[[r[1], s[1]]]),
                        {p[4], 1, Length[v]}, {p[3], 1, Length[v]},
                        {p[2], 1, Length[v]}] * f[[p[0], n, s[1], r[1]]], 0],
                {r[1], 1, Length[v]}, {s[1], 1, Length[v]},
                {p[0], 1, Length[v]}] +
            Sum[If[MemberQ[FOelems, {m, p[1], s[2], r[2]}],
                Sum[(v[[p[0], n]] * v[[p[1], p[0]]] * v[[p[4], s[2]]] * v[[r[2], p[3]]] *
                    l[[p[3], p[4]]]) / (w[[m, n]] * w[[m, p[0]]] *
                    w[[m, p[1]]] * w[[p[3], p[4]]] * w[[p[3], s[2]]] *
                    w[[r[2], s[2]]]) + (v[[p[0], n]] * v[[p[1], p[0]]] *
                    v[[p[3], s[2]]] * v[[r[2], p[4]]] * l[[p[4], p[3]]]) /
                    (w[[m, n]] * w[[m, p[0]]] * w[[m, p[1]]] *
                        w[[p[4], p[3]]] * w[[r[2], p[3]]] * w[[r[2], s[2]]]),
                    {p[4], 1, Length[v]}, {p[3], 1, Length[v]}] *
                    f[[m, p[1], s[2], r[2]]], 0] +
                If[MemberQ[FOelems, {p[0], p[1], s[2], r[2]}],
                    Sum[-((v[[m, p[0]]] * v[[p[1], n]] * v[[p[4], s[2]]] * v[[r[2], p[3]]] *
                        l[[p[3], p[4]]]) / (w[[m, n]] * w[[p[0], n]] *
                        w[[p[0], p[1]]] * w[[p[3], p[4]]] * w[[p[3], s[2]]] *
                        w[[r[2], s[2]]])) - (v[[m, p[0]]] * v[[p[1], n]] *
                        v[[p[3], s[2]]] * v[[r[2], p[4]]] * l[[p[4], p[3]]]) /
                        (w[[m, n]] * w[[p[0], n]] * w[[p[0], p[1]]] *
                            w[[p[4], p[3]]] * w[[r[2], p[3]]] * w[[r[2], s[2]]]),
                        {p[4], 1, Length[v]}, {p[3], 1, Length[v]}] *
                        f[[p[0], p[1], s[2], r[2]]], 0] +
                If[MemberQ[FOelems, {p[1], n, s[2], r[2]}],
                    Sum[(v[[m, p[0]]] * v[[p[0], p[1]]] * v[[p[4], s[2]]] * v[[r[2], p[3]]] *
                        l[[p[3], p[4]]]) / (w[[m, n]] * w[[p[0], n]] *
                        w[[p[1], n]] * w[[p[3], p[4]]] * w[[p[3], s[2]]] *
                        w[[r[2], s[2]]]) + (v[[m, p[0]]] * v[[p[0], p[1]]] *
                        v[[p[3], s[2]]] * v[[r[2], p[4]]] * l[[p[4], p[3]]]) /
                        (w[[m, n]] * w[[p[0], n]] * w[[p[1], n]] *
                            w[[p[4], p[3]]] * w[[r[2], p[3]]] * w[[r[2], s[2]]]),
                        {p[4], 1, Length[v]}, {p[3], 1, Length[v]}] *
                        f[[p[1], n, s[2], r[2]]], 0] +
                If[MemberQ[FOelems, {p[1], p[0], s[2], r[2]}],
                    Sum[-((v[[m, p[1]]] * v[[p[0], n]] * v[[p[4], s[2]]] * v[[r[2], p[3]]] *
                        l[[p[3], p[4]]]) / (w[[m, n]] * w[[m, p[0]]] *
                        w[[p[1], p[0]]] * w[[p[3], p[4]]] * w[[p[3], s[2]]] *
                        w[[r[2], s[2]]])) - (v[[m, p[1]]] * v[[p[0], n]] *
                        v[[p[3], s[2]]] * v[[r[2], p[4]]] * l[[p[4], p[3]]]) /
                        (w[[m, n]] * w[[m, p[0]]] * w[[p[1], p[0]]] *
                            w[[p[4], p[3]]] * w[[r[2], p[3]]] * w[[r[2], s[2]]]),
                        {p[4], 1, Length[v]}, {p[3], 1, Length[v]}] *
                        f[[p[1], p[0], s[2], r[2]]], 0], {r[2], 1, Length[v]},
                {s[2], 1, Length[v]}, {p[1], 1, Length[v]},
                {p[0], 1, Length[v]}] +
            Sum[-((I * v[[p[0], n]] * v[[p[1], p[0]]] * v[[p[2], p[1]]] * v[[p[3], p[2]]] *
                l[[m, p[3]]]) / (w[[m, n]] * w[[m, p[0]]] * w[[m, p[1]]] *
                w[[m, p[2]]] * w[[m, p[3]]])) +
                (I * v[[m, p[0]]] * v[[p[1], n]] * v[[p[2], p[1]]] * v[[p[3], p[2]]] *
                    l[[p[0], p[3]]]) / (w[[m, n]] * w[[p[0], n]] *
                    w[[p[0], p[1]]] * w[[p[0], p[2]]] * w[[p[0], p[3]]]) -
                (I * v[[m, p[0]]] * v[[p[0], p[1]]] * v[[p[2], n]] * v[[p[3], p[2]]] *
                    l[[p[1], p[3]]]) / (w[[m, n]] * w[[p[0], n]] * w[[p[1], n]] *
                    w[[p[1], p[2]]] * w[[p[1], p[3]]]) +
                (I * v[[m, p[1]]] * v[[p[0], n]] * v[[p[2], p[0]]] * v[[p[3], p[2]]] *
                    l[[p[1], p[3]]]) / (w[[m, n]] * w[[m, p[0]]] *
                    w[[p[1], p[0]]] * w[[p[1], p[2]]] * w[[p[1], p[3]]]) +
                (I * v[[m, p[0]]] * v[[p[0], p[1]]] * v[[p[1], p[2]]] * v[[p[3], n]] *
                    l[[p[2], p[3]]]) / (w[[m, n]] * w[[p[0], n]] * w[[p[1], n]] *
                    w[[p[2], n]] * w[[p[2], p[3]]]) -
                (I * v[[m, p[1]]] * v[[p[0], n]] * v[[p[1], p[2]]] * v[[p[3], p[0]]] *
                    l[[p[2], p[3]]]) / (w[[m, n]] * w[[m, p[0]]] *
                    w[[p[1], p[0]]] * w[[p[2], p[0]]] * w[[p[2], p[3]]]) +
                (I * v[[m, p[2]]] * v[[p[0], n]] * v[[p[1], p[0]]] * v[[p[3], p[1]]] *
                    l[[p[2], p[3]]]) / (w[[m, n]] * w[[m, p[0]]] * w[[m, p[1]]] *
                    w[[p[2], p[1]]] * w[[p[2], p[3]]]) -
                (I * v[[m, p[0]]] * v[[p[0], p[2]]] * v[[p[1], n]] * v[[p[3], p[1]]] *
                    l[[p[2], p[3]]]) / (w[[m, n]] * w[[p[0], n]] *
                    w[[p[0], p[1]]] * w[[p[2], p[1]]] * w[[p[2], p[3]]]) -
                (I * v[[m, p[0]]] * v[[p[0], p[1]]] * v[[p[1], p[2]]] * v[[p[2], p[3]]] *
                    l[[p[3], n]]) / (w[[m, n]] * w[[p[0], n]] * w[[p[1], n]] *
                    w[[p[2], n]] * w[[p[3], n]]) +
                (I * v[[m, p[1]]] * v[[p[0], n]] * v[[p[1], p[2]]] * v[[p[2], p[3]]] *
                    l[[p[3], p[0]]]) / (w[[m, n]] * w[[m, p[0]]] *
                    w[[p[1], p[0]]] * w[[p[2], p[0]]] * w[[p[3], p[0]]]) -
                (I * v[[m, p[2]]] * v[[p[0], n]] * v[[p[1], p[0]]] * v[[p[2], p[3]]] *
                    l[[p[3], p[1]]]) / (w[[m, n]] * w[[m, p[0]]] * w[[m, p[1]]] *
                    w[[p[2], p[1]]] * w[[p[3], p[1]]]) +
                (I * v[[m, p[0]]] * v[[p[0], p[2]]] * v[[p[1], n]] * v[[p[2], p[3]]] *
                    l[[p[3], p[1]]]) / (w[[m, n]] * w[[p[0], n]] *
                    w[[p[0], p[1]]] * w[[p[2], p[1]]] * w[[p[3], p[1]]]) +
                (I * v[[m, p[3]]] * v[[p[0], n]] * v[[p[1], p[0]]] * v[[p[2], p[1]]] *
                    l[[p[3], p[2]]]) / (w[[m, n]] * w[[m, p[0]]] * w[[m, p[1]]] *
                    w[[m, p[2]]] * w[[p[3], p[2]]]) -
                (I * v[[m, p[0]]] * v[[p[0], p[3]]] * v[[p[1], n]] * v[[p[2], p[1]]] *
                    l[[p[3], p[2]]]) / (w[[m, n]] * w[[p[0], n]] *
                    w[[p[0], p[1]]] * w[[p[0], p[2]]] * w[[p[3], p[2]]]) +
                (I * v[[m, p[0]]] * v[[p[0], p[1]]] * v[[p[1], p[3]]] * v[[p[2], n]] *
                    l[[p[3], p[2]]]) / (w[[m, n]] * w[[p[0], n]] * w[[p[1], n]] *
                    w[[p[1], p[2]]] * w[[p[3], p[2]]]) -
                (I * v[[m, p[1]]] * v[[p[0], n]] * v[[p[1], p[3]]] * v[[p[2], p[0]]] *
                    l[[p[3], p[2]]]) / (w[[m, n]] * w[[m, p[0]]] *
                    w[[p[1], p[0]]] * w[[p[1], p[2]]] * w[[p[3], p[2]]]),
                {p[3], 1, Length[v]}, {p[2], 1, Length[v]}, {p[1], 1, Length[v]},
                {p[0], 1, Length[v]}]];

SteadyStatePerturbativeSolve[v_?MatrixQ, l_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), {i_}] :=
    Table[SteadyStatePerturbativeSolve[v, l, w, f, {m, n}, {i}], {m, Length[v]}, {n, Length[v]}];

SteadyStatePerturbativeSolve[v_?MatrixQ, l_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), i_] :=
    Sum[SteadyStatePerturbativeSolve[v, l, w, f, {j}], {j, 0, i}];

SteadyStatePerturbativeSolve[v_?MatrixQ, l_?MatrixQ, w_?MatrixQ, {m_?IntegerQ, n_?IntegerQ}, {i_?IntegerQ}] :=
    SteadyStatePerturbativeSolve[v, l, w, ConstantArray[0, ConstantArray[Dimensions[v][[1]], 4]], {m, n}, {i}];

SteadyStatePerturbativeSolve[v_?MatrixQ, l_?MatrixQ, w_?MatrixQ, c : i_?IntegerQ | {i_?IntegerQ}] :=
    SteadyStatePerturbativeSolve[v, l, w, ConstantArray[0, ConstantArray[Dimensions[v][[1]], 4]], c];

TDPerturbativeSolve[v_?MatrixQ, r0_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), t1_, {m_, n_}, {0}] :=
    r0[[m, n]] / E^(I * t1 * w[[m, n]]);

TDPerturbativeSolve[v_?MatrixQ, r0_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), t1_, {m_, n_}, {1}] :=
    Sum[((v[[m, p]] * r0[[p, n]]) / w[[m, p]] - (E^(I * t1 * w[[m, p]]) * v[[m, p]] * r0[[p, n]]) / w[[m, p]] - (v[[p, n]] * r0[[m, p]]) / w[[p, n]] + (E^(I * t1 * w[[p, n]]) * v[[p, n]] * r0[[m, p]]) / w[[p, n]]) / E^(I * t1 * w[[m, n]]), {p, Length[v]}];

TDPerturbativeSolve[v_?MatrixQ, r0_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), t1_, {m_, n_}, {2}] :=
    Sum[If[m === r && n === s, 0, (((-I) * f[[m, n, s, r]] * v[[p2, s]] * v[[r, p1]] * r0[[p1, p2]]) / ((w[[m, n]] + w[[p2, p1]]) * w[[p2, s]] * (w[[p2, s]] + w[[r, p1]])) - (I * E^(I * (t1 * w[[m, n]] + t1 * w[[p1, r]])) * f[[m, n, s, r]] * v[[p1, s]] * v[[r, p2]] * r0[[p2, p1]]) / ((w[[m, n]] + w[[p1, r]]) * w[[p1, s]] * w[[r, p2]]) - (I * f[[m, n, s, r]] * v[[p1, s]] * v[[r, p2]] * r0[[p2, p1]]) / ((w[[m, n]] + w[[p1, p2]]) * w[[r, p2]] * (w[[p1, s]] + w[[r, p2]])) + I * ((f[[m, n, s, r]] * v[[p1, s]] * v[[r, p2]] * r0[[p2, p1]]) / ((w[[m, n]] + w[[p1, r]]) * w[[p1, s]] * w[[r, p2]]) + (E^(I * (t1 * w[[m, n]] + t1 * w[[p1, p2]])) * f[[m, n, s, r]] * v[[p1, s]] * v[[r, p2]] * r0[[p2, p1]]) / ((w[[m, n]] + w[[p1, p2]]) * w[[r, p2]] * (w[[p1, s]] + w[[r, p2]])) + (E^(I * t1 * w[[m, n]] - I * t1 * w[[r, s]]) * f[[m, n, s, r]] * v[[p1, s]] * v[[r, p2]] * r0[[p2, p1]]) / (w[[p1, s]] * w[[r, p2]] * (w[[m, n]] - w[[r, s]])) + (f[[m, n, s, r]] * v[[p1, s]] * v[[r, p2]] * r0[[p2, p1]]) / (w[[r, p2]] * (w[[p1, s]] + w[[r, p2]]) * (w[[m, n]] - w[[r, s]]))) - (I * f[[m, n, s, r]] * v[[p2, s]] * v[[r, p1]] * r0[[p1, p2]]) / (w[[p2, s]] * w[[r, p1]] * (w[[m, n]] - w[[r, s]])) - (I * E^(I * t1 * w[[m, n]] - I * t1 * w[[r, s]]) * f[[m, n, s, r]] * v[[p2, s]] * v[[r, p1]] * r0[[p1, p2]]) / (w[[p2, s]] * (w[[p2, s]] + w[[r, p1]]) * (w[[m, n]] - w[[r, s]])) - (I * f[[m, n, s, r]] * v[[p1, s]] * v[[r, p2]] * r0[[p2, p1]]) / (w[[p1, s]] * w[[r, p2]] * (w[[m, n]] - w[[r, s]])) - (I * E^(I * t1 * w[[m, n]] - I * t1 * w[[r, s]]) * f[[m, n, s, r]] * v[[p1, s]] * v[[r, p2]] * r0[[p2, p1]]) / (w[[r, p2]] * (w[[p1, s]] + w[[r, p2]]) * (w[[m, n]] - w[[r, s]])) - (I * E^(I * (t1 * w[[m, n]] + t1 * w[[s, p1]])) * f[[m, n, s, r]] * v[[p2, s]] * v[[r, p1]] * r0[[p1, p2]]) / (w[[p2, s]] * w[[r, p1]] * (w[[m, n]] + w[[s, p1]])) + I * ((E^(I * (t1 * w[[m, n]] + t1 * w[[p2, p1]])) * f[[m, n, s, r]] * v[[p2, s]] * v[[r, p1]] * r0[[p1, p2]]) / ((w[[m, n]] + w[[p2, p1]]) * w[[p2, s]] * (w[[p2, s]] + w[[r, p1]])) + (E^(I * t1 * w[[m, n]] - I * t1 * w[[r, s]]) * f[[m, n, s, r]] * v[[p2, s]] * v[[r, p1]] * r0[[p1, p2]]) / (w[[p2, s]] * w[[r, p1]] * (w[[m, n]] - w[[r, s]])) + (f[[m, n, s, r]] * v[[p2, s]] * v[[r, p1]] * r0[[p1, p2]]) / (w[[p2, s]] * (w[[p2, s]] + w[[r, p1]]) * (w[[m, n]] - w[[r, s]])) + (f[[m, n, s, r]] * v[[p2, s]] * v[[r, p1]] * r0[[p1, p2]]) / (w[[p2, s]] * w[[r, p1]] * (w[[m, n]] + w[[s, p1]])))) / E^(I * t1 * w[[m, n]])], {p2, Length[v]}, {p1, Length[v]}, {r, Length[v]}, {s, Length[v]}] + Sum[((E^(I * t1 * w[[p, n]]) * v[[m, p1]] * v[[p, n]] * r0[[p1, p]]) / (w[[m, p1]] * (w[[m, p1]] + w[[p, n]])) - (E^(I * (t1 * w[[m, p1]] + t1 * w[[p, n]])) * v[[m, p1]] * v[[p, n]] * r0[[p1, p]]) / (w[[m, p1]] * (w[[m, p1]] + w[[p, n]])) - (v[[m, p1]] * v[[p, n]] * r0[[p1, p]]) / (w[[p, n]] * (w[[m, p1]] + w[[p, n]])) + (E^(I * t1 * w[[p, n]]) * v[[m, p1]] * v[[p, n]] * r0[[p1, p]]) / (w[[p, n]] * (w[[m, p1]] + w[[p, n]])) + (v[[m, p]] * v[[p, p1]] * r0[[p1, n]]) / (w[[m, p]] * (w[[m, p]] + w[[p, p1]])) - (E^(I * t1 * w[[m, p]]) * v[[m, p]] * v[[p, p1]] * r0[[p1, n]]) / (w[[m, p]] * (w[[m, p]] + w[[p, p1]])) - (E^(I * t1 * w[[m, p]]) * v[[m, p]] * v[[p, p1]] * r0[[p1, n]]) / (w[[p, p1]] * (w[[m, p]] + w[[p, p1]])) + (E^(I * (t1 * w[[m, p]] + t1 * w[[p, p1]])) * v[[m, p]] * v[[p, p1]] * r0[[p1, n]]) / (w[[p, p1]] * (w[[m, p]] + w[[p, p1]])) - (v[[m, p]] * v[[p1, n]] * r0[[p, p1]]) / (w[[m, p]] * (w[[m, p]] + w[[p1, n]])) + (E^(I * t1 * w[[m, p]]) * v[[m, p]] * v[[p1, n]] * r0[[p, p1]]) / (w[[m, p]] * (w[[m, p]] + w[[p1, n]])) + (E^(I * t1 * w[[m, p]]) * v[[m, p]] * v[[p1, n]] * r0[[p, p1]]) / (w[[p1, n]] * (w[[m, p]] + w[[p1, n]])) - (E^(I * (t1 * w[[m, p]] + t1 * w[[p1, n]])) * v[[m, p]] * v[[p1, n]] * r0[[p, p1]]) / (w[[p1, n]] * (w[[m, p]] + w[[p1, n]])) + (v[[p, n]] * v[[p1, p]] * r0[[m, p1]]) / (w[[p, n]] * (w[[p, n]] + w[[p1, p]])) - (E^(I * t1 * w[[p, n]]) * v[[p, n]] * v[[p1, p]] * r0[[m, p1]]) / (w[[p, n]] * (w[[p, n]] + w[[p1, p]])) - (E^(I * t1 * w[[p, n]]) * v[[p, n]] * v[[p1, p]] * r0[[m, p1]]) / (w[[p1, p]] * (w[[p, n]] + w[[p1, p]])) + (E^(I * (t1 * w[[p, n]] + t1 * w[[p1, p]])) * v[[p, n]] * v[[p1, p]] * r0[[m, p1]]) / (w[[p1, p]] * (w[[p, n]] + w[[p1, p]]))) / E^(I * t1 * w[[m, n]]), {p1, Length[v]}, {p, Length[v]}];

TDPerturbativeSolve[v_?MatrixQ, r0_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), t1_, {i_}] :=
    Table[TDPerturbativeSolve[v, r0, w, f, t1, {m, n}, {i}], {m, Length[v]}, {n, Length[v]}];

TDPerturbativeSolve[v_?MatrixQ, r0_?MatrixQ, w_?MatrixQ, f_?(ArrayQ[#, 4]&), t1_, i_] :=
    Sum[TDPerturbativeSolve[v, r0, w, f, t1, {j}], {j, 0, i}];

End[];

EndPackage[];
