(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

BeginPackage["AtomicDensityMatrix`Common`", {"AtomicDensityMatrix`"}];

If[$VersionNumber < 7., PauliMatrix::usage = ""];

SampleExpression::usage = "";

$Sample::usage = "";

processGraphics::usage = "";

Graphics$::usage = "";

Show$::usage = "";

UnitTestInit::usage = "";

Begin["`Private`"];

HalfIntegerQ[x_] := IntegerQ[2x];

BlockDiagonal[m : {__?MatrixQ}] :=
    Module[{d = Dimensions /@ m},
        ArrayFlatten@Table[
            If[ i === j, m[[i]], Table[0, {d[[i, 1]]}, {d[[j, 2]]}]],
            {i, Length[m]}, {j, Length[m]}
        ]
    ];

Slope[{{x1_, y1_}, {x2_, y2_}}] := If[x2 != x1, (y2 - y1) / (x2 - x1), Infinity];

If[
    $VersionNumber < 7.,
    PauliMatrix[0] = {{1, 0}, {0, 1}};
    PauliMatrix[1] = {{0, 1}, {1, 0}};
    PauliMatrix[2] = {{0, -I}, {I, 0}};
    PauliMatrix[3] = {{1, 0}, {0, -1}};
    PauliMatrix[4] = {{1, 0}, {0, 1}};
    SetAttributes[PauliMatrix, Listable];
];

EnsureList[expr_] := If[ListQ@expr, expr, {expr}];

AllSameQ[list_, expr_] := And @@ (expr === # & /@ list);

ReplaceAt[expr_,
    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),
    pos : {},
    levelspec : (_Integer | Infinity | {_Integer | Infinity} | {(_Integer | Infinity), (_Integer | Infinity)}) : {0},
    opts : OptionsPattern[]
] := Replace[expr, Flatten[{rules}], levelspec, opts];

ReplaceAt[expr_,
    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),
    pos : ({__Integer} | _Integer),
    levelspec : (_Integer | Infinity | {_Integer | Infinity} | {(_Integer | Infinity), (_Integer | Infinity)}) : {0},
    opts : OptionsPattern[]
] := MapAt[Replace[#, Flatten[{rules}], levelspec, opts] &, expr, Flatten[{pos}]];

ReplaceAt[expr_,
    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),
    pos : {__List},
    levelspec : (_Integer | Infinity | {_Integer | Infinity} | {(_Integer | Infinity), (_Integer | Infinity)}) : {0},
    opts : OptionsPattern[]
] := Fold[ReplaceAt[#1, rules, #2, levelspec, opts] &, expr, pos];

ReplaceAllAt[expr_,
    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),
    pos : {}
] := ReplaceAll[expr, Flatten[{rules}]];

ReplaceAllAt[expr_,
    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),
    pos : ({__Integer} | _Integer)
] := MapAt[ReplaceAll[#, Flatten[{rules}]] &, expr, Flatten[{pos}]];

ReplaceAllAt[expr_,
    rules : ((_Rule | _RuleDelayed) | {(_Rule | _RuleDelayed) ...}),
    pos : {__List}
] := Fold[ReplaceAllAt[#1, rules, #2] &, expr, pos];

(* Used in MUnit tests to avoid tests failing due to differing round-off error on different machines *)
RationalizeMachineNumber[x_Real?MachineNumberQ, prec_] :=
    #1 10^#2 & @@ Round[MantissaExponent[x], 10^-prec];

RationalizeMachineNumber[x_Integer, prec_] := x;

RationalizeMachineNumbers[expr_, prec_ : 3] :=
    expr /. {
        x_Real?MachineNumberQ :> RationalizeMachineNumber[x, prec],
        x_Complex /; (MachineNumberQ[x] || MachineNumberQ[x / I]) :> RationalizeMachineNumber[Re@x, prec] + I RationalizeMachineNumber[Im@x, prec]
    };

processGraphics[expr_] :=
    Replace[expr /.
        {InterpolatingFunction[__] -> InterpolatingFunction, HoldPattern[DynamicModule[__]] -> Manipulate, Short[__] -> Short, Manipulate[__] -> Manipulate, Graphics3D[__] -> Graphics3D, _Image | _Graphics | _Graphics$ -> Graphics},
        x_ /; ByteCount@InputForm@x > 1048576/4 -> "BigCell"
    ];

SampleExpression[expression_] := Module[{pos, bytes, expr, pruned},
    expr = RationalizeMachineNumbers[expression /. a_SparseArray :> ArrayRules[a]];
    If[ByteCount@expr < 10000,
        expr,
        If[Length@expr > 1,
            expr = Sort[List @@ expr, ByteCount@#1 > ByteCount@#2 &]
        ];
        bytes = ByteCount@First@expr;
        pruned = Join[
            {First@expr},
            TakeWhile[Rest@expr, (bytes += ByteCount@#) < 10000&]
        ];
        $Sample[
            SampleExpression /@ pruned,
            Dimensions@expr,
            Hash@expr
        ]
    ]
];

UnitTestInit[] := (
    SetOptions[DensityMatrix,TimeDependence->True,Representation->"Zeeman",DMSymbol->\[Rho],DMLabel->None,ComplexExpandVariables->False,TimeVariable->t,AllowedElements->All,NeglectedElements->None];
    SetOptions[PlotAndFit, Method -> "Noninteractive"]
);

genericOptionsPattern[1] = _?(MatchQ[Flatten[{#}], {(_Rule | _RuleDelayed) ...}] &);

genericOptionsPattern[] = ___?(MatchQ[Flatten[{#}], {(_Rule | _RuleDelayed) ...}] &);

ZeroMatrix[n_Integer] := ConstantArray[0, {n, n}];
ZeroMatrix[n_Integer, SparseArray] := SparseArray[{}, {n, n}];

End[];

EndPackage[];
