(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`ExperimentalSystem`", {"AtomicDensityMatrix`"}];

Begin["`Private`"];

ExperimentalRegion[label_, {rules___Rule}] := ExperimentalRegion[label, rules];

RegionLabel[reg_ExperimentalRegion] := First@reg;
RegionLabel[sys : {_ExperimentalRegion..}] := RegionLabel /@ sys;

RegionParameters[reg_ExperimentalRegion] ^:= Rest[List @@ reg];

MapOnParameters[f_, reg_ExperimentalRegion] := ExperimentalRegion[RegionLabel@reg, f@RegionParameters@reg];

RegionParameterPattern[] = OpticalParameters | ElectricField | MagneticField | TransitRate[_] | Density;

RegionParameterQ[expr_] := MatchQ[expr, RegionParameterPattern[]];

ExperimentalSystemQ[expr_] := MatchQ[expr, {_ExperimentalRegion..}];

DefinedQ[region_ExperimentalRegion, x : RegionParameterPattern[]] := (x /. RegionParameters@region) =!= x;

DefinedQ[region_ExperimentalRegion, x : {RegionParameterPattern[]..}] := And @@ (DefinedQ[region, #]& /@ x);

DefinedQ[sys : {_ExperimentalRegion..}, x_] := And @@ (DefinedQ[#, x]& /@ sys);

DeleteParameter[region_ExperimentalRegion, x : RegionParameterPattern[]] := MapOnParameters[DeleteCases[#, x -> _]&, region];
DeleteParameter[region_ExperimentalRegion, x : {RegionParameterPattern[]..}] := Fold[DeleteParameter, region, x];
DeleteParameter[sys : {_ExperimentalRegion..}, x_] := DeleteParameter[#, x]& /@ sys;

AddParameter[region_ExperimentalRegion, r : Rule[x : RegionParameterPattern[], value_]] := MapOnParameters[Append[#, r]&, DeleteParameter[region, x]];

With[{f = #},
    f[reg_ExperimentalRegion] := f@reg = f /. RegionParameters@reg /. f -> f@RegionLabel@reg;
    f[sys : {_ExperimentalRegion..}] := f /@ sys;
]& /@ {ElectricField, MagneticField, OpticalParameters, Density};

TransitRate[reg1_ExperimentalRegion][reg2_ExperimentalRegion] := TransitRate[RegionLabel@reg1][reg2];

TransitRate[label_][region_ExperimentalRegion] :=
    Module[{value},
        If[
            RegionLabel@region === label,
            0,
            If[
                (value = TransitRate@label /. RegionParameters@region) =!= TransitRate@label,
                value,
                TransitRate[RegionLabel@region, label]
            ]
        ]
    ];

SelectRegions[es : {_ExperimentalRegion..}, crit__Rule] := Select[es, critFunc@crit];

SelectRegions[es : {_ExperimentalRegion..}, crit__Rule, n_?IntegerQ] := Take[Select[es, critFunc@crit], n];

SelectRegion[es : {_ExperimentalRegion..}, crit__Rule] := First@SelectStates[es, crit];

DeleteRegions[es : {_ExperimentalRegion..}, crit__Rule] := DeleteCases[es, _?(critFunc@crit)];

RegionPosition[es : {_ExperimentalRegion..}, crit__Rule] := Position[es, _?(critFunc@crit)];

End[];

EndPackage[];
