(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

BeginPackage["ProcessOptions`"];

OptionProcessor::usage = "";
OptionValidator::usage = "";
ProcessOption::usage = "";
HoldApply::usage = "";
EvaluateHold::usage = "";
ReleaseInnerHold::usage = "";
ProcessOptions::usage = "";

Begin["`Private`"];

OptionProcessor[_] = Identity;
OptionValidator[_] = _;

ProcessOption[f_, opt_, val_] := With[
    {processed = OptionProcessor[opt][val]},
    If[MatchQ[HoldComplete@processed, OptionValidator@opt] || MatchQ[processed, OptionValidator@opt],
        Unevaluated@processed,
        Message[f::adminvopt, opt, val];
        $Failed
    ]
];

HoldApply[f_, expr_Hold] := ReplacePart[expr, {1, 0} :> f];
EvaluateHold[expr_] := Hold[expr];
ReleaseInnerHold[expr_] := Replace[expr, (Hold | HoldForm | HoldPattern | HoldComplete)[x_] :> x, Infinity];

SetAttributes[ProcessOptions, HoldFirst];
ProcessOptions[def_] := Module[{vars, opt, opts, nvars, optvars, heldDef = Hold[def], function, body, condition, lastExpression},
    function = heldDef[[1, 1, 0]];
    If[function === Condition, function = heldDef[[1, 1, 1, 0]]];
    opts = Options[function][[All, 1]];
    If[heldDef[[1, 2, 0]] === Module,
        vars = Extract[heldDef, {1, 2, 1}, Hold];
        nvars = vars /. Hold[list_List] :> Length@Unevaluated@list;
        (* Add call to ProcessOption to all local variables that are the same as option names up to case. *)
        optvars = Last@Reap@Do[
            If[vars[[1, i, 0]] === Symbol &&
                Length[opt = Select[opts, StringMatchQ[ToString@#, ToString@vars[[1, i]], IgnoreCase -> True]&]] > 0,
                With[{var = Sow@vars[[1, i]], opt = First@opt, function = function},
                    heldDef = ReplacePart[heldDef, {1, 2, 1, i} :> (var = ProcessOption[function, opt, OptionValue@opt])]
                ]
            ],
            {i, nvars}
        ];
        If[Length@optvars > 0,
        (* Add condition that none of the ProcessOption calls returned $Failed *)
            With[{body = Extract[heldDef, {1, 2, 2}, Hold], optvars = optvars},
                heldDef = ReplacePart[heldDef, {1, 2, 2} :> Condition[body, FreeQ[optvars, $Failed]]]
            ]
        ]
    ];
    (* Expand calls to FilterRules and pass default options to called function *)
    With[{function = function},
        heldDef = heldDef /. {
            HoldPattern[subfunction_[args1___, FilterRules[opts_], args2___]] :> subfunction[args1, FilterRules[{opts, Options@function}, Options@subfunction], args2],
            HoldPattern[subfunction_[args1___, f_@FilterRules[opts_], args2___]] :> subfunction[args1, f@FilterRules[{opts, Options@function}, Options@subfunction], args2]
        }
    ];
    ReleaseHold@ReleaseInnerHold@heldDef
];

End[];

EndPackage[];
